% given F--the right hand side, J--Jacobian of the system 
% of linear equations, and an initial solution (omega, lambda, z, y),
% this part use line search to solve the system of equation. 
% User can choose Armijo line search or Wolfe line search.

% blck: number of total blocks.
% N:    vector, each element stores the dimension of corresponding block.
% m:    number of constrains.
% X:    type of primal. 'o', 'i', or 'b'. 
% S:    type of dual. 
%       S has one more type than X: 's' which means strict complementarity.
% lambda: norm of primal*sqrt(2); initially, it stores scale range of primal.
% omega:  norm of dual*sqrt(2); initially, it stores scale range of dual.
% ScaleA: scale range of linear constraint matrix A.
% Scaley: scale range of y.  Later it is used as a temporary variable to store
%  perturbation range of b, c, and A.

% M:    vector, each element stores the starting position of the corresponding
%       block.
  M(1)=1;
  for i = 2 : blck+1;
    M(i) = M(i-1)+N(i-1);
  end

  dmsn = M(blck+1)-1;  % total dimension of primal variable.
 
% starting column of lambda, z, y respectively. jlast is the dimension of J.
  jlambda = 2*blck+1;
  jz = jlambda + 2*blck;
  jy = jz+dmsn-blck;
  jlast = jy+m-1;

% a matrix's condtion number exceeds cndlimit will be considered ill-conditioned%  or singular. 
  mu = sqrt(jlast*eps);
  cndlimit = 1/sqrt(eps);

  rp = eps^(1/3)*2;

 % Calculate lambda, omega, z from x, s:
  lambda=[1;0]; omega=[1;0]; z=[1;0]; % ensure all are column vectors. 
  for i = 1: blck
     t = norm(s(M(i)+1:M(i+1)-1));
     omega(i,1) = (s(M(i))-t)/2;
     omega(i,2) = (s(M(i))+t)/2;
     t = norm(x(M(i)+1:M(i+1)-1));
     lambda(i,1) = (x(M(i))+t)/2;
     lambda(i,2) = (x(M(i))-t)/2;
     if t ~= 0
        z(M(i)-i+1:M(i+1)-i-1) = x(M(i)+1:M(i+1)-1)/t;
     elseif omega(i,1) ~= omega(i,2) 
        z(M(i)-i+1:M(i+1)-i-1) = s(M(i)+1:M(i+1)-1)/(omega(i,1)-omega(i,2));
     else
        z(M(i)-i+1) = 1;
        if N(i) > 2
          z(M(i)-i+2:M(i+1)-i-1) = 0;
        end
     end
  end
 
%perturb starting point
 inind = zeros(blck, 1);
 for i = 1: blck
    while lambda(i, 1) == omega(i, 1)
        lambda(i, 1) = lambda(i, 1) + rp*(rand-0.5);
        inind(i) = inind(i) + 1;
    end
    while lambda(i, 2) == omega(i, 2)
        lambda(i, 2) = lambda(i, 2) + rp*(rand-0.5);
        inind(i) = inind(i) + 1;
    end
 end
 fprintf('initial perturbation %d \n ', inind);



  st = A'*y-c;
  cost = 0;
  for i = 1: 17
     cost = cost + norm(st(3*i-1:3*i));
  end

  costopt = cost; yopt = y; kopt = 0;

% evaluate function value -F stored at F; Jacobian stored at J.
  Fvalue
  Jvalue
  feval = 1; % counter for function evaluation.

% Number of function values to be calculated for nonmonotone line search.
  endpsiv = 0;

% calculate Newton's step deltaw. 

   Psi0 = F'*F/2;  % corresponding unconstrained minimization problem. compare value.

   f00 = norm(F, inf); % for calculation of average initial gap.
   psiopt = Psi0; % store best results.
   kopt = 0; % store the number of best iteration.
   Psi = Psi0;  % current value.
   tol = eps^(1/3);  % tolerance, accuracy of solution.
   itnlimit = 100;
   minstep = eps^(2/3);
   maxstep = 1.0e+3; 
   beta = 0.5; % factor to reduce step size for Armijo rule.
   sigma = 1.0e-4;
   maxc = 10; % back track compare number for line search.
   fstamnb = 3; % Number of Armijo test before nonmonotone line search.
   fprintf('k = 0, cost = %2.10f, Psi=%e, ||F||_inf=%e \n' , cost, Psi, norm(F, inf))
   fprintf('Accuracy is %e, max number of iterations is %d, \n', tol, itnlimit)

   r = (1-rp)*beta;
   ndt = 0;

   k = 0;  % number of iteration
   Dvaluen

   omega0 = omega;  lambda0 = lambda;  z0 = z;  y0 = y;  F0 = F; chgw = norm(deltaw, inf); 

% linesch = input('Choose the line search rule [a: Armijo line search; w: Wolfe line search; n: nonmonotone linesearch (with perturbation) ]:', 's');
 if linesch == 'a' 
     % Armijo stepsize rule.
   while (norm(F, inf) >= tol) & (k < itnlimit) & (chgw > minstep)
     k = k+1;
     stepsize = 1;  % step size.
     omega = omega0+domega; 
     lambda = lambda0+dlambda;
     z = z0+deltaw(jz:jy-1);
     y = y0+deltaw(jy:jlast);  

     Fvalue
     feval = feval+1;
     amn = 0;  % Armijo iteration number
     dPsi = -F0'*J*deltaw;  % decrease of linear model with stepsize as unknown of Psi.    
     Psi = F'*F/2;

      while (Psi - Psi0) > (sigma*stepsize*dPsi) & (stepsize*chgw > minstep) 
        stepsize = beta * stepsize;
        omega = omega0+stepsize*domega; 
        lambda = lambda0+stepsize*dlambda;
        z = z0+stepsize*deltaw(jz:jy-1);
        y = y0+stepsize*deltaw(jy:jlast);  
        Fvalue;
        feval = feval+1;
        Psi = F'*F/2;
        amn = amn+1;
      end

      fprintf('k=%d amn=%d Psi=%e\n', k, amn, Psi);
      fprintf('dual infeasibility=%e, primal infeasibility=%e, gap=%e \n', norm(F(1:dmsn)), norm(F(dmsn+1:dmsn+m)), norm(F(dmsn+m+blck+1:jlast)) )

      % If stepsize is too small, the code get stuck, output the result.
      if Psi-Psi0 > sigma*stepsize*dPsi
        fprintf('Stop because the stepsize is %e, too small at iteration %d. \n', stepsize*chgw, k')
        chgw = 0; 

      else
        omega0 = omega;  lambda0 = lambda;  z0 = z;  y0 = y;  F0 = F;
        Jvalue
        Dvalue
        chgw = norm(deltaw, inf);
        Psi0 = Psi; 
      end
    end

  psiopt = Psi; kopt = k;
% End of Armijo line search.

 elseif linesch == 'w'

%	 Wolfe line search: Step selection by backtracking.
% J.E.Dennis,Jr. & Robert B.Schanbel, Numerical methods for unconstrained 
% optimization and nonlinear equations. Prentice Hall, 1983. p126-129.
% Only implement Armijo rule "because the backtracking strategy avoids 
% excessively small steps.  see also 
% Shultz, G. A., R. B. Schnabel, and R. H. Byrd (1982), " A family of trust 
% region based algorithms for unconstrained minimization with strong global
%  convergence properties", Univ. Colorado Computer Science TR CU-CS216-82.
   while (norm(F, inf) >= tol) & (k < itnlimit) & (chgw > minstep)
     k = k+1;
     stepsize = 1;  % step size.
     omega = omega0+domega; 
     lambda = lambda0+dlambda;
     z = z0+deltaw(jz:jy-1);
     y = y0+deltaw(jy:jlast);  
     Fvalue
     feval = feval+1;
     amn = 0;  % Armijo iteration number
     dPsi = -F0'*J*deltaw;  % decrease of linear model with lambda as unknown of Psi.    
     Psi = F'*F/2;
      if (Psi - Psi0) > (sigma*stepsize*dPsi) & (stepsize*chgw > minstep) 
        stepsize1 = stepsize; 
        Psi1 = Psi;

	% quadratic model.
        stepsize = -dPsi / (2*(Psi-Psi0-dPsi));
        if stepsize < 0.1
          stepsize = 0.1;
        end
        omega = omega0+stepsize*domega; 
        lambda = lambda0+stepsize*dlambda;
        z = z0+stepsize*deltaw(jz:jy-1);
        y = y0+stepsize*deltaw(jy:jlast);  
        Fvalue;
        feval = feval+1;
        Psi = F'*F/2;

	% Cubic model.
        while (Psi - Psi0) > (sigma*stepsize*dPsi) & (stepsize*chgw > minstep) 
          steppre = 1/(stepsize-stepsize1) * [1/(stepsize*stepsize), -1/stepsize1*stepsize1; -stepsize1/(stepsize*stepsize), stepsize/(stepsize1*stepsize1)] * [Psi-Psi0-dPsi*stepsize; Psi1-Psi0-dPsi*stepsize1];
           stepsize1 = stepsize;
           stepsize = (-steppre(2) + sqrt(steppre(2)*steppre(2)-3*steppre(1)*dPsi)) /(3*steppre(1));
           stepsize = real(stepsize);	
           if stepsize > stepsize1/2
             stepsize = stepsize1/2;
           end
             
	   if stepsize < stepsize1/10
             stepsize = stepsize1/10;
           end
           omega = omega0+stepsize*domega; 
           lambda = lambda0+stepsize*dlambda;
           z = z0+stepsize*deltaw(jz:jy-1);
           y = y0+stepsize*deltaw(jy:jlast);  
           Fvalue;
           feval = feval+1;
           Psi1 = Psi;
           Psi = F'*F/2;
         end
      end

       fprintf('k=%d amn=%d Psi=%e\n', k, amn, Psi);
       fprintf('dual infeasibility=%e, primal infeasibility=%e, gap=%e \n', norm(F(1:dmsn)), norm(F(dmsn+1:dmsn+m)), norm(F(dmsn+m+blck+1:jlast)) )

        omega0 = omega;  lambda0 = lambda;  z0 = z;  y0 = y;  F0 = F;
        Jvalue
        Dvalue
        Psi0 = Psi; 
        chgw = norm(deltaw, inf);
      end

  psiopt = Psi; kopt = k;
% End of wolfe line search.

 elseif linesch == 'n'
     % Nonmonotone line search.
   while (norm(F, inf) >= tol) & (k < itnlimit) & (chgw > minstep)
     k = k+1;
     stepsize = 1;  % step size.
     omega = omega0+domega; 
     lambda = lambda0+dlambda;
     z = z0+deltaw(jz:jy-1);
     y = y0+deltaw(jy:jlast);  

   nd = 0; %logic variable.  =1 if the iterate is nondiff., =0 otherwise. 
   % verify whether the iterate is differentiable or not.
    for i = 1: blck
      if lambda(i, 1) == omega(i, 1) | lambda(i, 2) == omega(i, 2)
         nd = 1;
      end
    end

  stepsizer = stepsize; % intermediate variable for use in the perturbation.

  % perturb stepsize when the iterate is nondiff.
   while nd == 1
      ndt = ndt + 1;
      stepsizer = (rand*(1-r) + r) * stepsize;
      omega = omega0 + stepsizer*domega;
      lambda = lambda0 + stepsizer * dlambda;
      z = z0 + stepsizer*deltaw(jz:jy-1);
      y = y0 + stepsizer*deltaw(jy:jlast);
      nd = 0;
      % verify whether the iterate is differentiable or not.
       for i = 1: blck
          if lambda(i, 1) == omega(i, 1) | lambda(i, 2) == omega(i, 2)
             nd = 1;
          end
       end
   end 

  stepsize = stepsizer;

     Fvalue
     feval = feval+1;
     amn = 0;  % Armijo iteration number
     dPsi = -F0'*J*deltaw;  % decrease of linear model with stepsize as unknown of Psi.    
     Psi = F'*F/2;

      while (Psi - Psi0) > (sigma*stepsize*dPsi) & (stepsize*chgw > minstep) 
        stepsize = beta * stepsize;
        omega = omega0+stepsize*domega; 
        lambda = lambda0+stepsize*dlambda;
        z = z0+stepsize*deltaw(jz:jy-1);
        y = y0+stepsize*deltaw(jy:jlast);  

   nd = 0; %logic variable.  =1 if the iterate is nondiff., =0 otherwise. 
   % verify whether the iterate is differentiable or not.
    for i = 1: blck
      if lambda(i, 1) == omega(i, 1) | lambda(i, 2) == omega(i, 2)
         nd = 1;
      end
    end

  stepsizer = stepsize; % intermediate variable for use in the perturbation.

  % perturb stepsize when the iterate is nondiff.
   while nd == 1
      ndt = ndt + 1;
      stepsizer = (rand*(1-r) + r) * stepsize;
      omega = omega0 + stepsizer*domega;
      lambda = lambda0 + stepsizer * dlambda;
      z = z0 + stepsizer*deltaw(jz:jy-1);
      y = y0 + stepsizer*deltaw(jy:jlast);
      nd = 0;
      % verify whether the iterate is differentiable or not.
       for i = 1: blck
          if lambda(i, 1) == omega(i, 1) | lambda(i, 2) == omega(i, 2)
             nd = 1;
          end
       end
   end 

  stepsize = stepsizer;

        Fvalue;
        feval = feval+1;
        Psi = F'*F/2;
        amn = amn+1;
      end

      fprintf('dual infeasibility=%e, primal infeasibility=%e, gap=%e \n', norm(F(1:dmsn)), norm(F(dmsn+1:dmsn+m)), norm(F(dmsn+m+blck+1:jlast)) )

  st = A'*y-c;
  cost = 0;
  for i = 1: 17
     cost = cost + norm(st(3*i-1:3*i));
  end
  if cost < costopt
     costopt = cost, yopt = y, kopt = k;
  end
   
      fprintf('k=%d, cost=%2.10f, Psi=%e, ||F||_inf=%e, amn=%d \n', k, cost, Psi, norm(F, inf), amn);

      % If stepsize is too small, the code get stuck, output the result.
      if Psi-Psi0 > sigma*stepsize*dPsi
        fprintf('Stop because the stepsize is %e, too small at iteration %d. \n', stepsize*chgw, k')
        chgw = 0; 

      else
        omega0 = omega;  lambda0 = lambda;  z0 = z;  y0 = y;  F0 = F;
        Jvalue
        Dvaluen
        chgw = norm(deltaw, inf);
        if endpsiv >=maxc
          Psi0 = max(Psiv);
        else 
          Psi0 = max(Psiv(1:endpsiv)); 
        end

        if psiopt > Psi
           psiopt = Psi; kopt = k;
        end

      end
    end
% End of nonmonotone line search.


 else error('Invalid choice of line search rule.')
 end 

 % Result output

 for i = 1 : blck
    x(M(i)) = lambda(i,1) + lambda(i,2);
    x(M(i)+1: M(i+1)-1) = (lambda(i,1)-lambda(i,2))*z(M(i)-i+1:M(i+1)-1-i);
    s(M(i)) = omega(i,1) + omega(i,2);
    s(M(i)+1: M(i+1)-1) = (omega(i,1)-omega(i,2))*z(M(i)-i+1: M(i+1)-i-1);
  end

 if k == itnlimit
   fprintf('Exceed iteration number limit, total function evaluation is %d, F''F/2 is now %e. \n', feval, Psi)
 else 
   fprintf('After iteration %d, F''F/2 is now %e. \n', k, Psi)
 end

   fprintf('costopt = %2.10f, kopt = %d\n', costopt, kopt);
   fprintf('yopt = %2.10f\n', yopt);
